import .lnat
import .subst
import .ftable
open lnat
open subst

inductive is_val : exp -> Prop :=
| const : Pi c, is_val (exp.const c)
| lam : Pi x t e, is_val (exp.lam x t e)
| const_ap : Pi {e}, Pi c, is_val e -> is_val (exp.ap (exp.const c) e)

inductive steps_to : exp -> exp -> Prop :=
| stepl : Pi {e1 e1'}, Pi e2, steps_to e1 e1' -> steps_to (exp.ap e1 e2) (exp.ap e1' e2)
| stepr : Pi {e1 e2 e2'}, is_val e1 -> steps_to e2 e2' -> steps_to (exp.ap e1 e2) (exp.ap e1 e2')
| subst : Pi {x e1 e1' e2}, Pi t, is_val e2 -> is_subst e1 x e2 e1' -> steps_to (exp.ap (exp.lam x t e1) e2) e1'

namespace is_val
  theorem nat.canon : Pi {e}, exp.has_typ (table.empty) e typ.nat -> is_val e ->
                      e = (exp.const (const.z)) \/ exists e', is_val e' /\ e = exp.ap (exp.const (const.s)) e'
  | nat.canon (@exp.has_typ.const op typ.nat ?(table.empty) Hconst) Hval :=
      have P : op = const.z, from const.nat.canon Hconst,
      or.inl (eq.subst P rfl)
  | nat.canon (@exp.has_typ.ap ?(table.empty) t typ.nat (exp.const op) e2 (exp.has_typ.const ?(table.empty) Hconst) Htyp2)
              (is_val.const_ap op Hval2) :=
      have Pop : op = const.s, from and.elim_left (const.arr.canon Hconst),
      or.inr (exists.intro e2 (and.intro Hval2 (eq.subst Pop rfl)))

  theorem lam.canon : Pi {e t1 t2}, exp.has_typ (table.empty) e (typ.arr t1 t2) -> is_val e ->
                      e = (exp.const const.s) \/ exists x e', e = exp.lam x t1 e'
  | lam.canon (@exp.has_typ.lam ?(table.empty) g' x t1 t2 e' Hctx Htyp) (is_val.lam x t1 e') :=
      or.inr (exists.intro x (exists.intro e' rfl))
  | lam.canon (@exp.has_typ.const op (typ.arr t1 t2) ?(table.empty) Hconst) Hval :=
      have Pop : op = const.s, from and.elim_left (const.arr.canon Hconst),
      or.inl (eq.subst Pop rfl)
  | lam.canon (@exp.has_typ.ap ?(table.empty) t (typ.arr t1 t2) (exp.const op) e2 (exp.has_typ.const ?(table.empty) Hconst) Htyp2)
              (is_val.const_ap op Hval2) :=
      have Pbad : typ.arr t1 t2 = typ.nat, from and.elim_right (and.elim_right (const.arr.canon Hconst)),
      typ.no_confusion Pbad
end is_val

-- This causes lean to raise SIGSEGV. Likely a lean bug.
theorem val_not_steps : forall {e}, forall e', is_val e -> steps_to e e' -> false := sorry
-- | val_not_steps e' (is_val.const c) Hsteps :=
--     match Hsteps with end
-- | val_not_steps e' (is_val.lam x t e) Hsteps :=
--     match Hsteps with end
-- | val_not_steps (exp.ap e' e) (@is_val.const_ap e c Hval) (steps_to.stepl e Hbad) :=
--     have Hvalc : is_val (exp.const c), from is_val.const c,
--     have IH : not (steps_to (exp.const c) e'), from val_not_steps e' Hvalc,
--     absurd Hbad IH
-- | val_not_steps (exp.ap (exp.const c) e') (@is_val.const_ap e c Hvale) (steps_to.stepr Hvalc Hbad) :=
--     have IH : not (steps_to e e'), from val_not_steps e' Hvale,
--     absurd Hbad IH

theorem progress : forall {t e}, exp.has_typ (table.empty) e t -> (exists e', steps_to e e') \/ is_val e
| progress (@exp.has_typ.var ?(table.empty) x t Hctx) :=
    absurd Hctx (table.empty_not_contains x t)
| progress (@exp.has_typ.const op t ?(table.empty) Hconst) := or.inr (is_val.const op)
| progress (@exp.has_typ.lam ?(table.empty) g' x _t t' _e Hctx Htyp) :=
    or.inr (is_val.lam x _t _e)
| progress (@exp.has_typ.ap ?(table.empty) _t t e1 e2 H1 H2) :=
    have IH1 : (exists e1', steps_to e1 e1') \/ is_val e1, from progress H1,
    have IH2 : (exists e2', steps_to e2 e2') \/ is_val e2, from progress H2,
    or.elim IH1
    ( assume Hsteps : exists e1', steps_to e1 e1',
      obtain e1' Hsteps, from Hsteps,
      or.inl (exists.intro (exp.ap e1' e2) (steps_to.stepl e2 Hsteps)) )
    ( assume Hval1 : is_val e1,
      or.elim IH2
      ( assume Hsteps : exists e2', steps_to e2 e2',
        obtain e2' Hsteps, from Hsteps,
        or.inl (exists.intro (exp.ap e1 e2') (steps_to.stepr Hval1 Hsteps)) )
      ( assume Hval2 : is_val e2,
        or.elim (is_val.lam.canon H1 Hval1)
        ( assume H : e1 = exp.const const.s,
          have Palmost : is_val (exp.ap (exp.const const.s) e2), from is_val.const_ap const.s Hval2,
          or.inr (eq.subst (eq.symm H) Palmost) )
        ( assume Hlam1 : exists x e', e1 = exp.lam x _t e',
          obtain x e' Heq1, from Hlam1,
          obtain e'' Hsubst, from exists_subst e' x e2,
          have P : steps_to (exp.ap (exp.lam x _t e') e2) e'', from steps_to.subst _t Hval2 Hsubst,
          or.inl (exists.intro e'' (eq.subst (eq.symm Heq1) P)) ) ) )

namespace aequiv
  theorem sound : forall {e1 e2 e1' e2'}, aequiv e1 e2 -> steps_to e1 e1' -> steps_to e2 e2' -> aequiv e1' e2'
  | sound (aequiv.id (exp.ap e1 e2)) (@steps_to.stepl e1 e1' e2 Hsteps1) (@steps_to.stepl e1 e1'' e2 Hsteps2) :=
      have Heq : aequiv e1 e1, from aequiv.id e1,
      have IH : aequiv e1' e1'', from sound Heq Hsteps1 Hsteps2,
      aequiv.ap IH (aequiv.id e2)
  | sound x y z := sorry
end aequiv

