import data.option
import logic.quantifiers

namespace table
  definition table (A B : Type) := A -> option B

  definition contains_key {A B : Type} (t : table A B) (k : A) : Prop :=
    not (t k = option.none)

  definition contains {A B : Type} (t : table A B) (k : A) (v : B) : Prop :=
    t k = option.some v

  definition ext {A B : Type} {t t' : table A B} : (forall k v, contains t k v <-> contains t' k v) -> t = t' :=
    assume H : forall k v, contains t k v <-> contains t' k v,
    have P : forall k, t k = t' k, from
      take k,
      match t k with
      | option.some v :=
          assume Heq : t k = option.some v,
          have Heq' : t' k = option.some v, from iff.elim_left (H k v) Heq,
          eq.subst (eq.symm Heq') Heq
      | option.none :=
          assume Heq : t k = option.none,
          match t' k with
          | option.some v :=
              assume Heq' : t' k = option.some v,
              have Heq : t k = option.some v, from iff.elim_right (H k v) Heq',
              eq.subst (eq.symm Heq') Heq
          | option.none :=
              assume Heq' : t' k = option.none,
              eq.subst (eq.symm Heq') Heq
          end rfl
      end rfl,
    funext P

  namespace contains
    theorem contains_key {A B : Type} (t : table A B) (k : A) (v : B) : contains t k v -> contains_key t k :=
      assume H : contains t k v,
      have H : t k = option.some v, from H,
      show not (t k = option.none), from
        assume Hbad : t k = option.none,
        have P : option.none = option.some v, from eq.subst Hbad H,
        absurd P (option.none_ne_some v)

    theorem inj {A B : Type} {t : table A B} {k : A} {v v' : B} : contains t k v -> contains t k v' -> v = v' :=
      assume H1 : t k = option.some v,
      assume H2 : t k = option.some v',
      option.some.inj (eq.subst H1 H2)
  end contains

  definition is_unionr {A B : Type} (l : table A B) (r : table A B) (t : table A B) : Prop :=
    forall k, forall v,
      contains t k v <-> contains r k v \/ (contains l k v /\ not (contains_key r k))

  namespace is_unionr
    theorem inj {A B : Type} {l r t t' : table A B} : is_unionr l r t -> is_unionr l r t' -> t = t' :=
      assume H1 : forall k v, contains t k v <-> contains r k v \/ (contains l k v /\ not (contains_key r k)),
      assume H2 : forall k v, contains t' k v <-> contains r k v \/ (contains l k v /\ not (contains_key r k)),
      have P : forall k v, contains t k v <-> contains t' k v, from
        take k v,
        iff.trans (H1 k v) (iff.symm (H2 k v)),
      ext P
  end is_unionr

  definition empty {A B : Type} : table A B :=
    fun x, option.none

  theorem empty_not_contains {A B : Type} (k : A) (v : B) : not (contains empty k v) :=
    assume Hbad : empty k = option.some v,
    have Hgood : empty k = option.none, from rfl,
    absurd (eq.subst Hgood Hbad) (option.none_ne_some v)

  definition singleton {A B : Type} (k : A) (v : B) [H : decidable_eq A] : table A B :=
    fun x, if x = k then option.some v else option.none

  theorem singleton_contains {A B : Type} (k : A) (v : B) [H : decidable_eq A] : contains (singleton k v) k v :=
  if_pos rfl

  definition unionr [reducible] {A B : Type} (l : table A B) (r : table A B) : table A B := fun k,
    match r k with
    | option.some v := option.some v
    | option.none := l k
    end

  theorem unionr_is_unionr {A B : Type} (l : table A B) (r : table A B) : is_unionr l r (unionr l r) :=
  fun k, fun v, iff.intro
    ( assume H : contains (unionr l r) k v,
      have Eq : unionr l r k = option.some v, from H,
      have R : r k = r k -> contains r k v \/ (contains l k v /\ not (contains_key r k)),
        from option.rec_on (r k)
        ( assume H : r k = option.none,
          have L : l k = option.some v, from calc
            l k = unionr l r k : eq.symm (eq.subst H rfl)
            ... = option.some v : Eq,
          have R : not (contains_key r k), from non_contradictory_intro H,
          or.inr (and.intro L R))
        ( take b : B,
          assume H : r k = option.some b,
          have L : r k = option.some v, from calc
            r k = option.some b : H
            ... = unionr l r k : eq.symm (eq.subst H rfl)
            ... = option.some v : Eq,
          or.inl L),
      R rfl)
    ( assume H, or.elim H
      ( assume H : contains r k v,
        have Eq : r k = option.some v, from H,
        show contains (unionr l r) k v, from eq.subst H rfl)
      ( assume H : contains l k v /\ not (contains_key r k),
        have Hl : contains l k v, from and.elim_left H,
        have Hr : not (contains_key r k), from and.elim_right H,
        have Pnone : r k = option.none, from option.rec_on (r k)
          ( assume H : r k = option.none, H )
          ( take b,
            assume H : r k = option.some b,
            have bad : not (option.none = r k), from eq.subst (eq.symm H) (option.none_ne_some b),
            absurd (ne.symm bad) Hr )
          rfl,
        have Eq : unionr l r k = l k, from eq.subst Pnone rfl,
        show contains (unionr l r) k v, from eq.subst Hl Eq))

  definition find [reducible] {A B : Type} (t : table A B) (k : A) : option B := t k

  theorem find_contains {A B : Type} (t : table A B) (k : A) (v : B) : find t k = option.some v <-> contains t k v :=
  iff.intro
    ( assume H : find t k = option.some v,
      show t k = option.some v, from eq.subst rfl H )
    ( assume H : t k = option.some v,
      show find t k = option.some v, from eq.subst H rfl )
end table
